----------------------------------------------------------------------------------
-- Company:             www.kampis-elektroecke.de
-- Engineer:            Daniel Kampert
-- 
-- Create Date:         02.08.2014 13:17:51
-- Design Name: 
-- Module Name:         VGA - VGA_Arch
-- Project Name: 
-- Target Devices:      XC7Z010CLG400-1
-- Tool Versions:       Vivado 2014.4
-- Description:         VGA Interface for a 640 x 480 Pixel Screen
-- 
-- Dependencies: 
-- 
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
-- 
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

entity VGA is
    Generic (
            Color_Width         : integer := 16;
            ROM_Address_Width   : integer := 11;
            ROM_Data_Width      : integer := 8;
            Char_Size           : integer := 8
            );
    Port (  Clock_VGA   : in STD_LOGIC;                                         -- Taktsignal fr VGA-Logik
            Reset       : in STD_LOGIC;                                         -- Reset fr VGA-Logik
            Mode        : in STD_LOGIC;                                         -- Auswahl Betriebsmodus, 0 -> Grafik, 1 -> Text
            Color       : in STD_LOGIC_VECTOR(15 downto 0);
            HSync       : out STD_LOGIC;                                        -- Ausgang fr H-Sync Signal
            VSync       : out STD_LOGIC;                                        -- Ausgang fr V-Sync Signal
            RGB         : out STD_LOGIC_VECTOR((Color_Width - 1) downto 0)      -- RGB Ausgang fr das VGA-Signal
            );
end VGA;

architecture VGA_Arch of VGA is

    signal Enable               : std_logic;

    -- Koordinaten des Zeigers auf dem Bildschirm
    signal Pixel_x_delay_0      : unsigned(9 downto 0);               
    signal Pixel_y_delay_0      : unsigned(9 downto 0); 
    signal Pixel_x_delay_1      : unsigned(9 downto 0);               
    signal Pixel_y_delay_1      : unsigned(9 downto 0); 
    signal Pixel_x_delay_2      : unsigned(9 downto 0);               
    signal Pixel_y_delay_2      : unsigned(9 downto 0); 
    signal Pixel_x_delay_3      : unsigned(9 downto 0);               
    signal Pixel_y_delay_3      : unsigned(9 downto 0); 
    
    signal HSync_delay_0        : std_logic;
    signal VSync_delay_0        : std_logic;
    signal HSync_delay_1        : std_logic;
    signal VSync_delay_1        : std_logic; 
    signal HSync_delay_2        : std_logic;
    signal VSync_delay_2        : std_logic; 
    signal HSync_delay_3        : std_logic;
    signal VSync_delay_3        : std_logic; 
    signal HSync_delay_4        : std_logic;
    signal VSync_delay_4        : std_logic; 
                            
    -- Font ROM
    signal ROM_Data     : std_logic_vector(0 to (ROM_Data_Width - 1));
    signal ROM_Address  : std_logic_vector(0 to (ROM_Address_Width - 1));
    signal ROM_Bit      : std_logic;
    signal Row_Addr     : std_logic_vector(2 downto 0);
    signal Col_Addr     : std_logic_vector(2 downto 0);
       
    -- Display RAM
    signal Display_Address  : unsigned(12 downto 0) := (others => '0');
    signal Zeichen_Address  : unsigned(6 downto 0);
    signal Offset           : unsigned(12 downto 0) := (others => '0');
    signal Character        : std_logic_vector(7 downto 0); 

    -- Einbinden des VGA-Controllers
    component VGA_Timing is
        Port (  HSync : out STD_LOGIC;
                VSync : out STD_LOGIC;
                Clock_VGA : in STD_LOGIC;
                Reset : in STD_LOGIC;
                x_out : out UNSIGNED(9 downto 0);                
                y_out : out UNSIGNED(9 downto 0) 
                );
    end component VGA_Timing;  
    
    -- Font-ROM fr VGA-Controller
    component Font_ROM is
        port (  ROM_Address : in STD_LOGIC_VECTOR (10 downto 0);
                ROM_Clock : in STD_LOGIC;
                ROM_DataOut : out STD_LOGIC_VECTOR (7 downto 0)
                );
    end component Font_ROM;
    
    -- Display RAM fr das Bild    
    component Display_RAM
        Generic (
                Addr_Width : integer := 13;
                Data_Width : integer := 8
                );
        Port (  Clock : in STD_LOGIC;
                Read_Write : in STD_LOGIC;
                Address : in STD_LOGIC_VECTOR((Addr_Width - 1) downto 0);
                Data_Out : out STD_LOGIC_VECTOR ((Data_Width - 1) downto 0);
                Data_In : in STD_LOGIC_VECTOR ((Data_Width - 1) downto 0)
                );
    end component;

begin

    HSync <= HSync_delay_4;
    VSync <= VSync_delay_4;

    VGA_Ctrl        : VGA_Timing port map (HSync_delay_0, VSync_delay_0, Clock_VGA, Reset, Pixel_x_delay_0, Pixel_y_delay_0);   
    Char_Font       : Font_ROM port map (ROM_Address, Clock_VGA, ROM_Data);
    Display         : Display_RAM port map (Clock_VGA, '1', std_logic_vector(Display_Address), Character, "00000000"); 

    --Farbausgabe    
    process(Clock_VGA)
    begin
        if(rising_edge(Clock_VGA)) then 
            if((ROM_Bit = '1')) then
                RGB <= Color;
            else
                RGB <= x"0000";
            end if;   
        end if;
    end process;
              
    -- ROM auslesen
    Row_Addr    <= std_logic_vector(Pixel_y_delay_2(2 downto 0));
    ROM_Address <= Character & Row_Addr;
    Col_Addr    <= std_logic_vector(Pixel_x_delay_2(2 downto 0));
    ROM_Bit     <= ROM_Data(to_integer(unsigned(Col_Addr)));     
    
    Zeichen_Address <= Pixel_x_delay_0(9 downto 3);  
    Display_Address <= Offset + Zeichen_Address;     

    process(Clock_VGA)
    begin 
        if(rising_edge(Clock_VGA)) then  
            if Pixel_x_delay_0 (2 downto 0)="000" then
                if Pixel_x_delay_0 =0 and  Pixel_y_delay_0 =0 then
                    Offset <=to_unsigned(0, Offset'length);
                elsif Pixel_x_delay_0 =0 and Pixel_y_delay_0(2 downto 0) =0 then 
                    Offset <= Offset + 80; 
                end if;        
            end if;   
        end if;
     end process;
        
    -- Signale um einen Takt verzgern
    process(Clock_VGA)
    begin
        if(rising_edge(Clock_VGA)) then
            HSync_delay_1 <= HSync_delay_0;
            VSync_delay_1 <= VSync_delay_0;
            HSync_delay_2 <= HSync_delay_1;
            VSync_delay_2 <= VSync_delay_1;
            HSync_delay_3 <= HSync_delay_2;
            VSync_delay_3 <= VSync_delay_2;
            HSync_delay_4 <= HSync_delay_3;
            VSync_delay_4 <= VSync_delay_3;
                        
            Pixel_x_delay_1 <= Pixel_x_delay_0;
            Pixel_y_delay_1 <= Pixel_y_delay_0;
            Pixel_x_delay_2 <= Pixel_x_delay_1;
            Pixel_y_delay_2 <= Pixel_y_delay_1;
            Pixel_x_delay_3 <= Pixel_x_delay_2;
            Pixel_y_delay_3 <= Pixel_y_delay_2;
        end if;
    end process;

end VGA_Arch;